﻿Imports System.IO
Imports System.Windows.Forms

Public Class Business
    Protected Shared BackupFolderName As String
    Protected Shared TextLog As Boolean
    Protected Shared LogFileName As String
    Protected Shared Settings As BOBSettings

    Public Shared Sub Main(ByVal SettingsFile As String)
        Dim ErrorMessage = String.Empty

        Try
            'If settings file does not exist, create it with defaults so can be edited in notepad
            If Not File.Exists(SettingsFile) Then BOBSettings.Save(SettingsFile, New BOBSettings)
            Settings = BOBSettings.Load(SettingsFile)

            'Following special directories are supported in the SourceFolder or BackupFolder paths
            '{AllUsersApplicationData}, {CurrentUserApplicationData}, {Desktop}, {MyDocuments}                
            '{MyMusic}, {MyPictures}, {ProgramFiles}, {Programs}, {Temp}                       
            Dim SourceFolder = Business.ParseSettingFolder(Settings.SourceFolder)
            Dim BackupFolder = Business.ParseSettingFolder(Settings.BackupFolder)
            Dim FullBackupDue = False

            'Wrap extension list in semicolons so any extension can be searched in 
            'format ";ext;" even if first or last in the list
            Dim Extensions = ";" + Settings.Extensions.Trim + ";"
            TextLog = Settings.TextLog

            'Create destination folder name, Backup on Build (BOB) + current date time
            BackupFolderName = "BOB_" + FileFriendlyDateTime(Now)
            LogFileName = Path.Combine(BackupFolder, BackupFolderName + "_Log.txt")

            If (Directory.Exists(SourceFolder)) Then
                Dim CreatedBackupFolder As Boolean = False

                'Create the backup folder from settings if not present
                If (Not Directory.Exists(BackupFolder)) Then
                    Directory.CreateDirectory(BackupFolder)
                    If TextLog Then WriteToLog("Created Primary Backup Folder: " + BackupFolder)
                End If

                'When was the last full backup?
                Dim LastFullBackup = GetLastFullBackupDate(BackupFolder)
                If TextLog Then WriteToLog("LastFullBackup: " + LastFullBackup.ToShortDateString)

                'Is a full backup due?
                If (Settings.FullBackupSchedule.Trim.Length > 0) Then
                    FullBackupDue = Business.IsFullBackupDue(LastFullBackup, Now, Settings.FullBackupSchedule)
                    If TextLog Then WriteToLog("FullBackupDue: " + FullBackupDue.ToString)
                End If

                'Perform the backup with recursive folder backups
                Dim SkippedExtensions As New List(Of String)
                Business.Backup(
                                SourceFolder,
                                BackupFolder,
                                FullBackupDue,
                                Settings.Extensions,
                                TextLog,
                                Settings.CompressBackup,
                                Settings.SkippedExtensionReport,
                                SkippedExtensions)


                'Log that a full backup was completed today if had been run
                If (FullBackupDue) Then
                    PutLastFullBackupDate(BackupFolder, Now.Date)
                    If TextLog Then WriteToLog("LastFullBackup: " + Date.Now.ToShortDateString)
                End If

                'Notify the user if the settings indicate to do so
                If Settings.CompletedNotification Then
                    Dim msg = "BuildOnBackup Completed Successfully"
                    MessageBox.Show(msg, "BackupOnBuild")
                End If

            Else
                ErrorMessage = String.Format(
                        "BackupOnBuild:Business:Main: SourceFolder of '{0}' does not exist",
                        Settings.SourceFolder
                        )
                If TextLog Then WriteToLog(ErrorMessage)
                MessageBox.Show(ErrorMessage, "BackupOnBuild")
            End If
        Catch ex As Exception
            ErrorMessage = String.Format("BackupOnBuild:Business:Main:Error: {0}", ex.Message)
            If Settings.TextLog Then WriteToLog(ErrorMessage)
            Throw New Exception("BackupOnBuild:Business:Main: " + ErrorMessage, ex)
        End Try

    End Sub

    ''' <summary>
    ''' Returns the full path represented by the application setting after expanding any special folder path
    ''' and resolving any relative path
    ''' </summary>
    ''' <param name="SettingFolder">Folder name to be parsed</param>
    ''' <returns>Parsed absolute path</returns>
    ''' <remarks>
    ''' Special Directores are supported in the input parameter if enclosed in {} and placed at the start of the parameter. 
    ''' The path is case-insensitive.
    ''' </remarks>
    Public Shared Function ParseSettingFolder(ByVal SettingFolder As String) As String
        Dim ParsedFolder = String.Empty

        'AllUsersApplicationData    Gets a path name pointing to the Application Data directory for the all users. 
        'CurrentUserApplicationData Gets a path name pointing to the Application Data directory for the current user. 
        'Desktop                    Gets a path name pointing to the Desktop directory. 
        'MyDocuments                Gets a path name pointing to the My Documents directory. 
        'MyMusic                    Gets a path name pointing to the My Music directory. 
        'MyPictures                 Gets a path name pointing to the My Pictures directory. 
        'ProgramFiles               Gets a path pointing to the Program Files directory. 
        'Programs                   Gets a path name pointing to the Programs directory. 
        'Temp                       Gets a path name pointing to the Temp directory. 

        Try

            Dim Folder = SettingFolder.Trim.ToLower

            'Is there a special folder in the path?  Must be at the start ...
            If (Folder.StartsWith("{")) Then
                Dim endBracket = Folder.IndexOf("}")
                If (endBracket > 1) Then
                    Dim specialFolder = Folder.Substring(0, endBracket + 1).Replace(" ", "")
                    Select Case specialFolder
                        Case "{AllUsersApplicationData}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.AllUsersApplicationData
                        Case "{CurrentUserApplicationData}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.CurrentUserApplicationData
                        Case "{Desktop}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.Desktop
                        Case "{MyDocuments}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.MyDocuments
                        Case "{MyMusic}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.MyMusic
                        Case "{MyPictures}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.MyPictures
                        Case "{ProgramFiles}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.ProgramFiles
                        Case "{Programs}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.Programs
                        Case "{Temp}".ToLower
                            ParsedFolder = My.Computer.FileSystem.SpecialDirectories.Temp
                        Case Else
                            Throw New Exception("BackupOnBuild:Business:ParseSettingFolder: Invalid SetttingFolder value of " + SettingFolder)
                    End Select

                    If (Not SettingFolder.EndsWith("}")) Then
                        Dim remainingSetting = SettingFolder.Substring(endBracket + 1)
                        If remainingSetting.StartsWith("\") AndAlso remainingSetting.Length > 1 Then
                            remainingSetting = remainingSetting.Substring(1)
                        End If
                        ParsedFolder = Path.Combine(ParsedFolder, remainingSetting)
                    End If

                End If
            Else
                ParsedFolder = Path.GetFullPath(Folder)
            End If

        Catch ex As Exception
            Throw New Exception("BackupOnBuild:Business:ParseSettingFolder:Error: " + ex.Message, ex)
        End Try

        Return ParsedFolder
    End Function

    ''' <summary>
    ''' Determine if a full backup is due based on the date of the last full backup, the backup schedule, and today's date
    ''' </summary>
    ''' <param name="LastFullBackup">Date of the last full backup as recorded in the BOB_LastFullBackupDate.txt file</param>
    ''' <param name="FullBackupSchedule"></param>
    ''' <returns>True is a full backup is due else False</returns>
    ''' <remarks></remarks>
    Public Shared Function IsFullBackupDue(ByVal LastFullBackup As DateTime, ByVal AsOfDate As DateTime, ByVal FullBackupSchedule As String) As Boolean
        'Day of Week enumeration has values Sunday=0, Monday=1, etc.

        Dim FullBackupDue = False

        Try
            Dim NextFull = LastFullBackup
            Dim LastFullBackupDay = LastFullBackup.DayOfWeek
            Dim LastBackupDOW = DirectCast(LastFullBackup.DayOfWeek, Integer)

            'Number of days to wait for next full backup for each target date
            'Indexed by Day of Week of the last full backup
            Dim SunInterval = {7, 6, 5, 4, 3, 2, 1}
            Dim MonInterval = {1, 7, 6, 5, 4, 3, 2}
            Dim TueInterval = {2, 1, 7, 6, 5, 4, 3}
            Dim WedInterval = {3, 2, 1, 7, 6, 5, 4}
            Dim ThuInterval = {4, 3, 2, 1, 7, 6, 5}
            Dim FriInterval = {5, 4, 3, 2, 1, 7, 6}
            Dim SatInterval = {6, 5, 4, 3, 2, 1, 7}

            'Determine number of days to wait from last full backup
            Select Case FullBackupSchedule.Trim.ToUpper
                Case "ALWAYS"
                    NextFull = DateTime.MinValue
                Case "NEVER"
                    NextFull = DateTime.MaxValue
                Case "MONTHLY"
                    NextFull = LastFullBackup.AddMonths(1)
                Case "WEEKLY"
                    NextFull = LastFullBackup.AddDays(7)
                Case "DAILY"
                    NextFull = LastFullBackup.AddDays(1)
                Case "HOURLY"
                    NextFull = LastFullBackup.AddHours(1)
                Case "SUNDAY"
                    NextFull = LastFullBackup.AddDays(SunInterval(LastBackupDOW))
                Case "MONDAY"
                    NextFull = LastFullBackup.AddDays(MonInterval(LastBackupDOW))
                Case "TUESDAY"
                    NextFull = LastFullBackup.AddDays(TueInterval(LastBackupDOW))
                Case "WEDNESDAY"
                    NextFull = LastFullBackup.AddDays(WedInterval(LastBackupDOW))
                Case "THURSDAY"
                    NextFull = LastFullBackup.AddDays(ThuInterval(LastBackupDOW))
                Case "FRIDAY"
                    NextFull = LastFullBackup.AddDays(FriInterval(LastBackupDOW))
                Case "SATURDAY"
                    NextFull = LastFullBackup.AddDays(SatInterval(LastBackupDOW))
                Case "BIWEEEKLY"
                    NextFull = LastFullBackup.AddDays(14)
            End Select

            If (AsOfDate >= NextFull) Then FullBackupDue = True

        Catch ex As Exception
            Throw New Exception("BackupOnBuild:Business:IsFullBackupDue:Error: " + ex.Message, ex)
        End Try

        Return FullBackupDue
    End Function

    Public Shared Function Backup(ByVal SourceFolder As String,
                                    ByVal BackupFolder As String,
                                    ByVal FullBackupDue As Boolean,
                                    ByVal Extensions As String,
                                    ByVal TextLog As Boolean,
                                    ByVal CompressBackup As Boolean,
                                    ByVal SkippedExtensionsReport As Boolean,
                                    ByRef SkippedExtensions As List(Of String)) As Boolean
        Dim Successful = True

        Try

            Dim BackupPath = Path.Combine(BackupFolder, BackupFolderName)
            Directory.CreateDirectory(BackupPath)

            'Recursive method to backup a specific folder and its sub-folders
            FolderBackup(SourceFolder, BackupPath, FullBackupDue, Extensions, TextLog, SkippedExtensionsReport, SkippedExtensions)

            'Create the skipped extension report if indicated by the settings
            If SkippedExtensionsReport Then
                Dim FileName = BackupFolderName + "_SkippedExtensionReport.txt"
                Dim FullFileName = Path.Combine(BackupFolder, FileName)
                Dim sb As New System.Text.StringBuilder
                Dim Heading = "The following file extensions were not backed up " +
                    "since not present in backup extensions list:"
                sb.AppendLine(Heading)
                For Each item In SkippedExtensions
                    sb.AppendLine("." + item)
                Next
                File.WriteAllText(FullFileName, sb.ToString)
            End If

            'Zip the backup and delete the backup folders that were zipped, if indicated by the settings
            If CompressBackup Then
                'Compress the entire backup into a zip file
                Dim zip As New ICSharpCode.SharpZipLib.Zip.FastZip
                zip.CreateZip(BackupPath + ".zip", BackupPath, True, "")

                DeleteBackupFolderToRecyleBin(BackupPath)
            End If

        Catch ex As Exception
            Throw New Exception("BackupOnBuild:Business:Backup Error: " + ex.Message, ex)
        End Try

        Return Successful
    End Function

    Public Shared Sub FolderBackup(ByVal FullPathSourceFolder As String,
                                   ByVal FullPathBackupFolder As String,
                                   ByVal FullBackup As Boolean,
                                   ByVal Extensions As String,
                                   ByVal TextLog As Boolean,
                                   ByVal SkippedExtensionsReport As Boolean,
                                   ByRef SkippedExtensions As List(Of String))
        If TextLog Then
            WriteToLog("FolderBackup Called - " +
                                   "FullPathSourceFolder: " + FullPathSourceFolder + ", " +
                                   "FullPathBackupFolder: " + FullPathBackupFolder + ", " +
                                   "SkippedExtensions: " + SkippedExtensions.Log)
        End If

        Dim Files = Directory.GetFiles(FullPathSourceFolder)
        Dim Folders = Directory.GetDirectories(FullPathSourceFolder)

        If Not Directory.Exists(FullPathBackupFolder) Then
            Directory.CreateDirectory(FullPathBackupFolder)
            If TextLog Then WriteToLog("Created Directory: " + FullPathBackupFolder)
        End If

        For Each FileItem In Files
            Dim FileItemInfo = New FileInfo(FileItem)
            Dim FileExt = ""
            Dim CopyFile = False

            If (FullBackup) Then
                CopyFile = True
            Else
                FileExt = New FileInfo(FileItem).Extension.ToLower.Replace(".", "")
                If (Extensions.Contains(";" + FileExt + ";")) Then CopyFile = True
            End If
            If CopyFile Then
                Dim DestFileName = FullPathBackupFolder + "\" + FileItemInfo.Name
                File.Copy(FileItem, DestFileName)
                If TextLog Then WriteToLog("Copied File " + FileItem + " to " + DestFileName)
            Else
                'Should we add the extension to the Skipped Extensions list?
                If (SkippedExtensionsReport AndAlso SkippedExtensions.IndexOf(FileExt) = -1) Then
                    SkippedExtensions.Add(FileExt)
                    If TextLog Then WriteToLog("Added extension: " + FileExt + " From File: " + FileItem)
                End If
            End If
        Next

        For Each FolderItem In Folders
            Dim FolderItemInfo = New DirectoryInfo(FolderItem)
            Call FolderBackup(
                FullPathSourceFolder + "\" + FolderItemInfo.Name,
                FullPathBackupFolder + "\" + FolderItemInfo.Name,
                FullBackup,
                Extensions,
                TextLog,
                SkippedExtensionsReport,
                SkippedExtensions)
        Next
    End Sub

    ''' <summary>
    '''Returns a formatted datetime without any invalid filename characters
    ''' </summary>
    ''' <param name="dt">DateTime to be formatted</param>
    ''' <returns>File friendly formatted DateTime string</returns>
    ''' <remarks></remarks>
    Public Shared Function FileFriendlyDateTime(ByVal dt As DateTime) As String
        Dim FileFriendlyDT =
            dt.Month.ToString("00-") +
            dt.Day.ToString("00-") +
            dt.Year.ToString("0000") +
                                "_" +
            dt.Hour.ToString("00-") +
            dt.Minute.ToString("00-") +
            dt.Second.ToString("00")

        Return FileFriendlyDT
    End Function

    ''' <summary>
    ''' Immediately appends message to the log file
    ''' </summary>
    ''' <param name="message">Message to be written to the log file</param>
    ''' <remarks>Immediately writes message to log file 
    ''' to ensure latest status should application abort</remarks>
    Public Shared Sub WriteToLog(ByVal message As String)
        Dim LogTime = Now
        File.AppendAllText(LogFileName,
                           LogTime.ToShortDateString + " " +
                           LogTime.ToShortTimeString + ": " +
                           message + vbCrLf)
    End Sub

    Public Shared Sub DeleteBackupFolderToRecyleBin(ByVal FolderPathToDelete As String)
        Try
            My.Computer.FileSystem.DeleteDirectory(
                FolderPathToDelete,
                FileIO.UIOption.OnlyErrorDialogs,
                FileIO.RecycleOption.SendToRecycleBin)

            If TextLog Then WriteToLog("Deleted Path: " + FolderPathToDelete)

        Catch ex As PathTooLongException
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:Path Too Long: " +
                ex.Message + " : " + ex.StackTrace)

        Catch ex As ArgumentNullException
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:Path is nothing: " +
                ex.Message + " : " + ex.StackTrace)

        Catch ex As ArgumentException
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:Malformed Path: " +
                ex.Message + " : " + ex.StackTrace)

        Catch ex As DirectoryNotFoundException
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:Invalid Path: " +
                ex.Message + " : " + ex.StackTrace)

        Catch ex As IOException
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:File or Folder in Use: " +
                ex.Message + " : " + ex.StackTrace)

        Catch ex As NotSupportedException
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:Invalid colon character in path: " +
                ex.Message + " : " + ex.StackTrace)

        Catch ex As Exception
            If TextLog Then WriteToLog(
                "Business:DeleteBackupFolderToRecycleBin:Unhandled Exception: " +
                ex.Message + " : " + ex.StackTrace)
        End Try
    End Sub

    Public Shared Function GetLastFullBackupDate(ByVal BackupFolder As String) As Date
        Dim LastFullBackup As Date = Date.MinValue
        Dim LastFullBackupDateFileName = Path.Combine(BackupFolder, "BOB_LastFullBackupDate.txt")
        If File.Exists(LastFullBackupDateFileName) Then
            Dim LastFullBackupDateString As String() = File.ReadAllLines(LastFullBackupDateFileName).ToArray
            If (Not IsNothing(LastFullBackupDateString) AndAlso LastFullBackupDateString.Count > 0) Then
                Date.TryParse(LastFullBackupDateString(0), LastFullBackup)
            End If
        End If
        Return LastFullBackup
    End Function

    Public Shared Sub PutLastFullBackupDate(ByVal BackupFolder As String, ByVal LastFullBackup As Date)
        Dim LastFullBackupDateFileName = Path.Combine(BackupFolder, "BOB_LastFullBackupDate.txt")
        File.WriteAllText(LastFullBackupDateFileName, LastFullBackup.ToShortDateString + vbCrLf)
    End Sub
End Class
